SystemOrganization addCategory: #'NativeBoost-Unix'!

SharedPool subclass: #NBUnixConstants
	instanceVariableNames: ''
	classVariableNames: 'MAP_32BIT MAP_ANON MAP_ANONYMOUS MAP_DENYWRITE MAP_EXECUTABLE MAP_FAILED MAP_FILE MAP_FIXED MAP_GROWSDOWN MAP_LOCKED MAP_NONBLOCK MAP_NORESERVE MAP_POPULATE MAP_PRIVATE MAP_SHARED MAP_STACK MAP_TYPE PROT_EXEC PROT_GROWSDOWN PROT_GROWSUP PROT_NONE PROT_READ PROT_WRITE RTLD_BINDING_MASK RTLD_DEEPBIND RTLD_DEFAULT RTLD_GLOBAL RTLD_LAZY RTLD_LOCAL RTLD_NEXT RTLD_NODELETE RTLD_NOLOAD RTLD_NOW'
	poolDictionaries: ''
	category: 'NativeBoost-Unix'!

!NBUnixConstants class methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 9/25/2010 08:37'!
initDlopenFlags

"From <bits/dlfcn.h>"

" The MODE argument to `dlopen' contains one of the following: "
	RTLD_LAZY	 := 16r00001. "Lazy function call binding."
	RTLD_NOW := 16r00002. " Immediate function call binding. "
	RTLD_BINDING_MASK := 16r3. " Mask of binding time value. "
	RTLD_NOLOAD	:= 16r00004. " Do not load the object."
	RTLD_DEEPBIND := 16r00008. " Use deep binding. "

" If the following bit is set in the MODE argument to `dlopen',
   the symbols of the loaded object and its dependencies are made
   visible as if the object were linked directly into the program.  "
	RTLD_GLOBAL := 16r00100.

" Unix98 demands the following flag which is the inverse to RTLD_GLOBAL.
   The implementation does this by default and so we can define the
   value to zero.  "
	RTLD_LOCAL := 0.

" Do not delete object when closed.  "
	RTLD_NODELETE := 16r01000.
	
" From <dlfcn.h> "
" If the first argument of `dlsym' or `dlvsym' is set to RTLD_NEXT
   the run-time address of the symbol called NAME in the next shared
   object is returned.  The 'next' relation is defined by the order
   the shared objects were loaded.  "
	RTLD_NEXT := 16rFFFFFFFF.  "  ((void *) -1l) "
   
" If the first argument to `dlsym' or `dlvsym' is set to RTLD_DEFAULT
   the run-time address of the symbol called NAME in the global scope
   is returned.  " 
	RTLD_DEFAULT := 0. 
   

! !

!NBUnixConstants class methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 9/25/2010 09:12'!
initMmapFlags

	"From <bits/mman.h>"

	" Memory protection flags, can be bitored "	
		
	PROT_READ := 16r1.	" Page can be read.  "
	PROT_WRITE := 16r2.	" Page can be written. "
	PROT_EXEC := 16r4.	" Page can be executed. "
	PROT_NONE := 16r0. 	" Page can not be accessed. "

	PROT_GROWSDOWN  := 16r01000000. " Extend change to start of growsdown vma (mprotect only). "
	PROT_GROWSUP := 16r02000000. "Extend change to start of growsup vma (mprotect only). "

	
	" Sharing types (must choose one and only one of these). "
	MAP_SHARED := 16r01.		"Share changes."
	MAP_PRIVATE := 16r02.		"Changes are private."
	
	MAP_TYPE := 16r0f. 	"Mask for type of mapping. "

	" Other flags. "
	MAP_FIXED := 16r10.	"Interpret addr exactly. "

	MAP_FILE := 0.
	MAP_ANONYMOUS := 16r20. "Don't use a file."
	MAP_ANON := MAP_ANONYMOUS.
	MAP_32BIT := 16r40.	"Only give out 32-bit addresses."

	"These are Linux-specific."

	MAP_GROWSDOWN := 16r00100.	"Stack-like segment."
	MAP_DENYWRITE := 16r00800.		" ETXTBSY "
	MAP_EXECUTABLE := 16r01000.	" Mark it as an executable."
	MAP_LOCKED := 16r02000. 		" Lock the mapping. "
	MAP_NORESERVE := 16r04000.		" Don't check for reservations. "
	MAP_POPULATE := 16r08000.		" Populate (prefault) pagetables. "
	MAP_NONBLOCK := 16r10000.		" Do not block on IO."
	MAP_STACK := 16r20000.			" Allocation is for a stack. "

	MAP_FAILED := 16rFFFFFFFF.  "-1"! !

!NBUnixConstants class methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 9/25/2010 06:57'!
initialize
	" self initialize "
	"from <bits/mman.h>"
	self
		initDlopenFlags;
		initMmapFlags! !

NBExternalHeapManager subclass: #NBUnixExternalHeapManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'NBUnixConstants'
	category: 'NativeBoost-Unix'!

!NBUnixExternalHeapManager commentStamp: 'Igor.Stasenko 9/26/2010 04:05' prior: 0!
Unix heap implementation, based on mmap()/munmap()
functions to manage external heap.!

!NBUnixExternalHeapManager methodsFor: 'flags' stamp: 'Igor.Stasenko 9/25/2010 08:49'!
mapFlags

	^ MAP_ANON bitOr: MAP_PRIVATE! !

!NBUnixExternalHeapManager methodsFor: 'private' stamp: 'IgorStasenko 11/24/2012 17:24'!
mmapLength: bytesToAllocate prot: protFlags flags: mapFlags into: returnValueBuffer
		
	<primitive: #primitiveNativeCall module: #NativeBoostPlugin>

	"Note:  mmap using 32-bit offset, while mmap64 - 64bit one"

	^ self nbCallout 
		options: #( - optDirectProxyFnAddress );
		function: #(
			NBBootstrapUlong ( 0, uint bytesToAllocate, int protFlags, int mapFlags,
				-1,  0 "offset" )
		
		) emit:  [:gen | | mmap |
			mmap := NativeBoost forCurrentPlatform getGlobalSymbolPointer: 'mmap'.
			gen asm mov: mmap asUImm32 to: gen asm EAX.
			gen asm call: gen asm EAX
			]

 ! !

!NBUnixExternalHeapManager methodsFor: 'memory pages' stamp: 'Igor.Stasenko 9/26/2010 03:37'!
primAllocatePage: bytesToAllocate
	| ptr buf |

	buf := ByteArray new: 4.

	self 
		mmapLength: bytesToAllocate 
		prot: self protFlags 
		flags: self mapFlags 
		into: buf.

	ptr := buf unsignedLongAt: 1 bigEndian: false.
	ptr = MAP_FAILED ifTrue: [ self error: 'Unable to map virtual memory' ].
	
	^ ptr

 ! !

!NBUnixExternalHeapManager methodsFor: 'memory pages' stamp: 'Igor.Stasenko 9/26/2010 03:45'!
primFreePage: aMemoryPage

	"call munmap()"

	| res |
	res := self unmap: aMemoryPage address length: aMemoryPage length.
	res = 0 ifFalse: [ self error: 'error during unmapping virtual memory page' ]! !

!NBUnixExternalHeapManager methodsFor: 'flags' stamp: 'Igor.Stasenko 9/26/2010 03:51'!
protFlags
	"since NativeBoost using heap for placing executable code there,
	it is important that all allocated memory should be executable"
	^  (PROT_READ bitOr: PROT_WRITE) bitOr: PROT_EXEC! !

!NBUnixExternalHeapManager methodsFor: 'private' stamp: 'IgorStasenko 11/24/2012 17:25'!
unmap: addr  length: len 
		
	<primitive: #primitiveNativeCall module: #NativeBoostPlugin>

	"unmap memory page"

	^ self nbCallout 
		function: #( int munmap (uint addr, ulong  len)  ) 
		module: RTLD_DEFAULT
 ! !

NativeBoost subclass: #NativeBoostLinux32
	instanceVariableNames: 'heap'
	classVariableNames: ''
	poolDictionaries: 'NBUnixConstants'
	category: 'NativeBoost-Unix'!

!NativeBoostLinux32 class methodsFor: 'platform id' stamp: 'Igor.Stasenko 9/22/2010 03:12'!
targetPlatformId
	
	^ Linux32PlatformId ! !

!NativeBoostLinux32 methodsFor: 'retrieving symbols' stamp: 'Igor.Stasenko 9/26/2010 05:00'!
CLibrary
	"answer a C library path, or module handle,
	use in FFI callouts as module: argument to call C run-time library function(s)"
	
	"use global symbol space"
	^ RTLD_DEFAULT! !

!NativeBoostLinux32 methodsFor: 'retrieving symbols' stamp: 'JavierPimas 9/27/2011 14:46'!
VMModule
	"taken from CLibrary. Should work for fetching VM functions too.
	use in FFI callouts as module: argument to call C run-time library function(s)"
	
	"use global symbol space"
	^ RTLD_DEFAULT! !

!NativeBoostLinux32 methodsFor: 'memory operations' stamp: 'Igor.Stasenko 9/25/2010 09:37'!
basicAllocate: numBytes
	"Allocate a memory block with given size in bytes,
	answer an NBExternalAddress instance - address to the beginning of memory block"
	| addr |
	addr := heap allocate: numBytes.
	addr = 0 ifTrue: [ ^ nil ].
	
	^ NBExternalAddress value: addr! !

!NativeBoostLinux32 methodsFor: 'bootstrapping' stamp: 'IgorStasenko 11/24/2012 16:02'!
bootstrapGetGlobalSymbolPointer: symbolName into: returnValueBuffer
	" call  dlsym(RTLD_DEFAULT, symbolName) "

	<primitive: 'primitiveNativeCall' module: 'NativeBoostPlugin' error: errorCode>
	^ self nbCallout
		options: #( - optUseStackPointer );
		function: #( NBBootstrapUlong (RTLD_DEFAULT , String symbolName) )
		emit: [:gen |	 | asm |
				self assert: (self primitiveDlsym ~= 0).
				
				asm := gen asm.
				asm mov: (self primitiveDlsym asUImm) to: asm EAX.
				asm call: asm EAX.
		] 
	! !

!NativeBoostLinux32 methodsFor: 'memory operations' stamp: 'Igor.Stasenko 9/26/2010 04:13'!
free: address

	"Free the external memory, allocated using #allocate: message.
	Note: never pass pointers, which you allocated by other means"
	
	heap free: address value! !

!NativeBoostLinux32 methodsFor: 'retrieving symbols' stamp: 'IgorStasenko 1/24/2012 12:38'!
getGlobalSymbolPointer: symbolName 

	| bytes addr |
	bytes := ByteArray new: 8.
	self bootstrapGetGlobalSymbolPointer: symbolName into: bytes.
	addr := bytes unsignedLongAt: 1 bigEndian: false.
	
	addr = 0 ifTrue: [ self error: 'failed to get a symbol address: ', symbolName ].
	^ NBExternalAddress value: addr
	! !

!NativeBoostLinux32 methodsFor: 'bootstrapping' stamp: 'Igor.Stasenko 9/26/2010 03:50'!
initializeExternalHeap

	"initialize external memory heap"
		
	heap := NBUnixExternalHeapManager new.

	! !

!NativeBoostLinux32 methodsFor: 'retrieving symbols' stamp: 'IgorStasenko 1/18/2012 13:45'!
loadSymbol: aSymbolName fromModule: moduleName

	(moduleName = '' or: [moduleName = 0 ]) ifTrue: [
		"if module name is empty, try to retrieve symbol from globals "
		^ self getGlobalSymbolPointer: aSymbolName 
		].
	^ super loadSymbol: aSymbolName fromModule: moduleName! !

!NativeBoostLinux32 methodsFor: 'accessing' stamp: 'IgorStasenko 8/5/2011 08:48'!
newAssembler
	^ AJx86Assembler new.! !

!NativeBoostLinux32 methodsFor: 'accessing' stamp: 'Igor.Stasenko 9/22/2010 03:25'!
pointerSize
	^ 4! !

!NativeBoostLinux32 methodsFor: 'primitives' stamp: 'Igor.Stasenko 9/25/2010 06:31'!
primitiveDlopen
	"retrieve a pointer to dlopen(...) function"
	<primitive: 'primitiveDlopenFn' module: #NativeBoostPlugin>
		
	^ self primitiveFailed! !

!NativeBoostLinux32 methodsFor: 'primitives' stamp: 'Igor.Stasenko 9/25/2010 06:30'!
primitiveDlsym
	"retrieve a pointer to dlsym(...) function"
	<primitive: 'primitiveDlsymFn' module: #NativeBoostPlugin>
		
	^ self primitiveFailed! !

!NativeBoostLinux32 methodsFor: 'accessing' stamp: 'IgorStasenko 8/3/2011 08:08'!
stackAlignment

	^ 1! !
NBUnixConstants initialize!
